//------------------------------------------------------------------------------
// CERN BE-CO-HT
// LHC Instability Trigger Distribution (LIST)
// https://ohwr.org/projects/list
//------------------------------------------------------------------------------
//
// unit name:   main
//
// description: A SystemVerilog testbench to exercise all the main features of
// LIST
//
//------------------------------------------------------------------------------
// Copyright CERN 2018
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`include "wrtd_driver.svh"
`include "vhd_wishbone_master.svh"

`timescale 1ns/10fs

module simple_tdc_driver
  (
   input [3:0]	     addr,
   output reg [27:0] data,
   input	     wr,
   input	     rd,
   output reg	     ef1,
   output reg	     ef2
   );

   typedef struct    {
      int	     channel;
      time	     ts;
   } acam_fifo_entry;

   acam_fifo_entry fifo1[$], fifo2[$];

   task pulse(int channel, time ts);

      acam_fifo_entry ent;

      ent.channel = channel % 4;
      ent.ts = ts;

      if (channel >= 0 && channel <= 3)
	fifo1.push_back(ent);
      else
	fifo2.push_back(ent);

      #100ns;
      if(fifo1.size())
	ef1 = 0;
      if(fifo2.size())
	ef2 = 0;

   endtask // pulse

   initial begin
      ef1 = 1;
      ef2 = 1;
      data = 28'bz;
   end

   always@(negedge rd) begin
      if (addr == 8) begin
	 acam_fifo_entry ent;
	 ent=fifo1.pop_front();
	 data <= ent.ts | (ent.channel << 26) | (1<<17);

      end else if (addr == 9) begin
	 acam_fifo_entry ent;
	 ent=fifo2.pop_front();
	 data <= ent.ts | (ent.channel << 26) | (1<<17);

      end else
	data <= 28'bz;

      #10ns;

      ef1 <= (fifo1.size() ? 0 : 1);
      ef2 <= (fifo2.size() ? 0 : 1);

   end

endmodule // simple_tdc_driver

module simple_fdelay_mon
  (
   input [3:0] len,
   input [9:0] val,
   input [3:0] pulse
   );

   task display_pulse ();
      string tmp;
      tmp = $sformatf("[FDEL] Pulse: len=%x, val=%x, out=%x", len, val, pulse);
      $display (tmp);
   endtask // display_pulse

   always@(posedge pulse) begin
      display_pulse();
   end

endmodule // simple_fdelay_mon

module main;

   reg rst_n = 0;
   reg clk_125m_pllref = 0;
   reg clk_20m_vcxo = 0;
   reg clk_ext = 0;
   reg clk_acam = 0;

   wire a2btxp, a2btxn, b2atxp, b2atxn;

   wire tdc_ef1, tdc_ef2, tdc_rd_n;

   wire [3:0] tdc_addr;

   wire [27:0] tdc_data;

   wire [3:0]  fdl_len, fdl_pulse;

   wire [9:0]  fdl_val;

  always #50ns clk_ext        <= ~clk_ext;
  always #4ns clk_125m_pllref <= ~clk_125m_pllref;
  always #16ns clk_acam       <= ~clk_acam;
  always #20ns clk_20m_vcxo   <= ~clk_20m_vcxo;

   initial begin
      repeat(20) @(posedge clk_20m_vcxo);
      rst_n = 1;
   end

   const uint64_t mt_base = 'h2_0000;

   //
   // WRTD transmitter (TDC)
   //

   simple_tdc_driver
     TDC
       (
	.addr(tdc_addr),
	.data(tdc_data),
	.wr(1'b0),
	.rd(tdc_rd_n),
	.ef1(tdc_ef1),
	.ef2(tdc_ef2)
	);

   svec_list_top #
     (
      .g_simulation (1),
      .g_sim_bypass_vme (1),
      .g_wr_dpram_initf("../../../dependencies/wr-cores/bin/wrpc/wrc_phy8_sim.bram"),
      .g_mt_cpu0_initf("../../../software/firmware/tdc/wrtd-rt-tdc.bram"),
      .g_mt_cpu1_initf("none")
      )
   DUT_A
     (
      .rst_n_i                  (rst_n),
      .clk_125m_pllref_p_i      (clk_125m_pllref),
      .clk_125m_pllref_n_i      (~clk_125m_pllref),
      .clk_125m_gtp_p_i         (clk_125m_pllref),
      .clk_125m_gtp_n_i         (~clk_125m_pllref),
      .clk_20m_vcxo_i           (clk_20m_vcxo),
      .sim_wb_i                 (HostA.out),
      .sim_wb_o                 (HostA.in),
      .sfp_txp_o                (a2btxp),
      .sfp_txn_o                (a2btxn),
      .sfp_rxp_i                (b2atxp),
      .sfp_rxn_i                (b2atxn),
      .fmc0_tdc_pll_status_i    (1'b1),
      .fmc0_tdc_125m_clk_p_i    (clk_125m_pllref),
      .fmc0_tdc_125m_clk_n_i    (~clk_125m_pllref),
      .fmc0_tdc_acam_refclk_p_i (clk_acam),
      .fmc0_tdc_acam_refclk_n_i (~clk_acam),
      .fmc0_tdc_ef1_i           (tdc_ef1),
      .fmc0_tdc_ef2_i           (tdc_ef2),
      .fmc0_tdc_err_flag_i      (1'b0),
      .fmc0_tdc_int_flag_i      (1'b0),
      .fmc0_tdc_data_bus_io     (tdc_data),
      .fmc0_tdc_address_o       (tdc_addr),
      .fmc0_tdc_rd_n_o          (tdc_rd_n)
      );

   IVHDWishboneMaster HostA
     (
      .clk_i   (DUT_A.clk_sys_62m5),
      .rst_n_i (DUT_A.rst_sys_62m5_n));

   IMockTurtleIRQ IrqMonitorA (`MT_ATTACH_IRQ(DUT_A.cmp_mock_turtle));

   WrtdDriver drvA;

   initial begin

      @(posedge DUT_A.rst_sys_62m5_n);
      @(posedge DUT_A.clk_sys_62m5);

      drvA = new (HostA.get_accessor(), mt_base, IrqMonitorA, "DUT:A");

      drvA.init();

      forever begin
	 drvA.update ();
	 #1us;
      end

   end // initial begin DUTA

   //
   // WRTD receiver (Fine Delay)
   //

/* -----\/----- EXCLUDED -----\/-----
   simple_fdelay_mon
     FDL
       (
	.len   (fdl_len),
	.val   (fdl_val),
	.pulse (fdl_pulse)
	);
 -----/\----- EXCLUDED -----/\----- */

   svec_list_top #
     (
      .g_simulation (1),
      .g_sim_bypass_vme (1),
      .g_wr_dpram_initf("../../../dependencies/wr-cores/bin/wrpc/wrc_phy8_sim.bram"),
      .g_mt_cpu0_initf("none"),
      .g_mt_cpu1_initf("../../../software/firmware/fd/wrtd-rt-fd.bram")
      )
   DUT_B
     (
      .rst_n_i               (rst_n),
      .clk_125m_pllref_p_i   (clk_125m_pllref),
      .clk_125m_pllref_n_i   (~clk_125m_pllref),
      .clk_125m_gtp_p_i      (clk_125m_pllref),
      .clk_125m_gtp_n_i      (~clk_125m_pllref),
      .clk_20m_vcxo_i        (clk_20m_vcxo),
      .sim_wb_i              (HostB.out),
      .sim_wb_o              (HostB.in),
      .sfp_txp_o             (b2atxp),
      .sfp_txn_o             (b2atxn),
      .sfp_rxp_i             (a2btxp),
      .sfp_rxn_i             (a2btxn),
      .fmc1_fd_clk_ref_p_i   (clk_125m_pllref),
      .fmc1_fd_clk_ref_n_i   (~clk_125m_pllref),
      .fmc1_fd_delay_len_o   (fdl_len),
      .fmc1_fd_delay_val_o   (fdl_val),
      .fmc1_fd_delay_pulse_o (fdl_pulse)
      );

   IVHDWishboneMaster HostB
     (
      .clk_i   (DUT_B.clk_sys_62m5),
      .rst_n_i (DUT_B.rst_sys_62m5_n));

   IMockTurtleIRQ IrqMonitorB (`MT_ATTACH_IRQ(DUT_B.cmp_mock_turtle));

   WrtdDriver drvB;

   initial begin

      @(posedge DUT_B.rst_sys_62m5_n);
      @(posedge DUT_B.clk_sys_62m5);

      drvB = new (HostB.get_accessor(), mt_base, IrqMonitorB, "DUT:B");

      drvB.init();

      forever begin
	 drvB.update ();
	 #1us;
      end

   end // initial begin DUTB

   initial begin // others

      $timeformat (-6, 3, "us", 10);

      // Skip WR SoftPLL lock
      force DUT_A.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.U_SOFTPLL.U_Wrapped_Softpll.out_locked_o = 3'b111;
      force DUT_B.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.U_SOFTPLL.U_Wrapped_Softpll.out_locked_o = 3'b111;
      // Silence Xilinx unisim DSP48A1 warnings about invalid OPMODE
      force DUT_A.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D1.OPMODE_dly = 0;
      force DUT_A.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D2.OPMODE_dly = 0;
      force DUT_A.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D3.OPMODE_dly = 0;
      force DUT_B.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D1.OPMODE_dly = 0;
      force DUT_B.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D2.OPMODE_dly = 0;
      force DUT_B.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D3.OPMODE_dly = 0;

      #200us;

      forever begin
	 #200us;
	 TDC.pulse(0, 0);
	 TDC.pulse(1, 0);
	 TDC.pulse(2, 0);
      end

   end // initial begin others

endmodule // main
