/*
 * Copyright (C) 2013-2018 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include <errno.h>
#include <string.h>

#include "mockturtle-rt.h"
#include <mockturtle-framework.h>
#include "wrtd-common.h"
#include "wrtd-fw-common.h"
#include "hw/fmctdc-direct.h"
#include "hw/tdc_regs.h"
#include "loop-queue.h"

#include "wrtd-rt-tdc.h"

#define WRTD_IN_HMQ 0
#define WRTD_IN_RMQ 0

#ifdef SIMULATION
static const int is_simulation = 1;
#else
static const int is_simulation = 0;
#endif

enum wrtd_in_wr_link {
	WR_LINK_OFFLINE = 1,
	WR_LINK_ONLINE,
	WR_LINK_SYNCING,
	WR_LINK_SYNCED,
	WR_LINK_TDC_WAIT,
};

static struct wrtd_in wrtd_in_dev;
static struct wrtd_in_channel wrtd_in_channels[TDC_NUM_CHANNELS];

static uint32_t sent_packets = 0; /**< Total number of packets sent */
static int coalesce_count = 0; /**< RMQ coalescing counter */
static int wr_state = 0; /**< White-Rabbit link status */
static uint32_t tai_start = 0;


static inline int wr_link_up()
{
	return dp_readl(BASE_DP_TDC_REGS + TDC_REG_WR_STAT) & TDC_WR_STAT_LINK;
}

static inline int wr_time_locked()
{
	return dp_readl(BASE_DP_TDC_REGS + TDC_REG_WR_STAT) & TDC_WR_STAT_AUX_LOCKED;
}

static inline int wr_time_ready()
{
	return dp_readl(BASE_DP_TDC_REGS + TDC_REG_WR_STAT) & TDC_WR_STAT_TIME_VALID;
}

static inline int wr_is_timing_ok()
{
	return (wr_state == WR_LINK_SYNCED);
}


/**
 * It sends a log entry to the host system
 */
static int wrtd_in_trigger_log(int type, int miss_reason,
			       struct wrtd_in_channel *st,
			       struct wrtd_trigger_entry *ent)
{
	struct wrtd_log_entry *log;
	struct trtl_fw_msg msg;
	int ret;

	if (!(st->config.log_level & type))
		return -EINVAL;

	ret = mq_claim(TRTL_HMQ, WRTD_IN_HMQ);
	if (ret == -EBUSY)
		return ret;

	mq_map_out_message(TRTL_HMQ, WRTD_IN_HMQ, &msg);

	msg.header->msg_id = WRTD_IN_ACTION_LOG;
	msg.header->len = sizeof(struct wrtd_log_entry) / 4;
	log = (struct wrtd_log_entry *)msg.payload;
	log->type = type;
	log->channel = st ? st->n : -1;
	log->miss_reason = miss_reason;
	log->seq = ent->seq;
	log->id = ent->id;
	log->ts = ent->ts;

	pr_debug("%s: %d %d\n\r", __func__, type, miss_reason);

	mq_send(TRTL_HMQ, WRTD_IN_HMQ);

	return 0;
}


/**
 * Creates a trigger message with timestamp (ts) for the channel (ch) and
 * pushes it to the RMQ output (without sending) and sends immediately through
 * the loopback queue
 */
static inline void send_trigger (struct wrtd_trigger_entry *ent)
{
	volatile struct wrtd_trigger_message *msg;
	int ret;

	if (coalesce_count == 0) {
		ret = mq_claim(TRTL_RMQ, WRTD_IN_RMQ);
		if (ret < 0) {
			pr_error("RMQ full\n\r");
			return;
		}
	}

	msg = mq_map_out_buffer(TRTL_RMQ, WRTD_IN_RMQ);
	msg->triggers[coalesce_count].id = ent->id;
	msg->triggers[coalesce_count].seq = ent->seq;
	msg->triggers[coalesce_count].ts = ent->ts;

	loop_queue_push(&ent->id, ent->seq, &ent->ts);

	coalesce_count++;
}


/**
 * Flushes the triggers in the RMQ output buffer to the WR Network
 */
static inline void flush_tx ()
{
	volatile struct wrtd_trigger_message *msg = mq_map_out_buffer(TRTL_RMQ,
								      WRTD_IN_RMQ);

	/* Embed transmission time for latency measurement */
	msg->transmit_seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
	msg->transmit_cycles = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);
	msg->count = coalesce_count;

	mq_send(TRTL_RMQ, WRTD_IN_RMQ);
	coalesce_count = 0;
	sent_packets++;
}


/**
 * Processes a pulse with timestamp (ts) arriving on channel (channel)
 */
static inline void do_channel(int channel, struct wr_timestamp *ts)
{
	struct wrtd_in_channel *ch = &wrtd_in_channels[channel];
	struct wrtd_trigger_entry ent = {
		.ts = *ts,
	};

	/* Apply timebase offset to align TDC time with WR timebase */
	ts_sub(ts, &ch->config.timebase_offset);
	ch->stats.last_tagged = *ts;

	/* Log raw value if needed */
	wrtd_in_trigger_log(WRTD_LOG_RAW, 0, ch, &ent);

	ch->stats.total_pulses++;

	/* Apply trigger delay */
	ts_add(ts, &ch->config.delay);

	/* Enable/Arm/Trigger logic */
	if (!((ch->config.flags & WRTD_TRIGGER_ASSIGNED ) &&
	      (ch->config.flags & WRTD_ARMED)))
		return;

	if (!wr_is_timing_ok()) {
		ch->stats.miss_no_timing++;
		wrtd_in_trigger_log(WRTD_LOG_MISSED, WRTD_MISS_NO_WR, ch, &ent);
		return;
	}

	ent.ts = *ts;
	ent.id = ch->config.id;
	ent.seq = ch->stats.seq++;

	ch->config.flags |= WRTD_TRIGGERED;
	if (ch->config.mode == WRTD_TRIGGER_MODE_SINGLE )
		ch->config.flags &= ~WRTD_ARMED;

	ch->stats.sent_pulses++;
	ch->stats.last_sent = ent;

	send_trigger(&ent);
	wrtd_in_trigger_log(WRTD_LOG_SENT, 0, ch, &ent);

	ch->config.flags |= WRTD_LAST_VALID;
}

/**
 * Handles input timestamps from all TDC channels, calling do_output() on
 * incoming pulses
 */
static inline void do_input(void)
{
	int i;

	/* Prepare for message transmission */
	mq_claim(TRTL_RMQ, WRTD_IN_RMQ);

	/* We can send up to TDC_TRIGGER_COALESCE_LIMIT triggers in a
	   single message - the loop will iterate up to this limit or exit
	   immediately if there's no more input pulses in the TDC FIFO */
	for (i = 0; i < TDC_TRIGGER_COALESCE_LIMIT; i++) {
		uint32_t fifo_sr = dp_readl(BASE_DP_TDC_DIRECT + DR_REG_FIFO_CSR);
		struct wr_timestamp ts;
		int meta;

		/* Poll the FIFO and read the timestamp */
		if(fifo_sr & DR_FIFO_CSR_EMPTY)
			break;

		ts.seconds = dp_readl(BASE_DP_TDC_DIRECT + DR_REG_FIFO_R0);
		ts.ticks = dp_readl(BASE_DP_TDC_DIRECT + DR_REG_FIFO_R1);
		meta = dp_readl(BASE_DP_TDC_DIRECT + DR_REG_FIFO_R2);

		/* Convert from ACAM bins (81ps) to WR time format. Numerical
		   hack used  to avoid time-consuming division. */
		ts.frac = ( (meta & 0x3ffff) * 5308 ) >> 7;
		ts.ticks += ts.frac >> 12;
		ts.frac &= 0xfff;

		/* Make sure there's no overflow after conversion */
		if (ts.ticks >= 125000000) {
			ts.ticks -= 125000000;
			ts.seconds ++;
		}

		int channel = (meta >> 19) & 0x7;

		/* Pass the timestamp to triggering/TX logic */
		do_channel(channel, &ts);
	}

	/* Flush the RMQ buffer if it contains anything */
	if(coalesce_count)
		flush_tx();
}


/**
 * It generate a software trigger accorging to the trigger entry coming
 * from the user space. The user space entry must provide the trigger
 * delay, then this function will add it to the current time to compute
 * the correct fire time.
 */
static int wrtd_in_trigger_sw(struct trtl_fw_msg *msg_i, struct trtl_fw_msg *msg_o)
{
	struct wrtd_trigger_entry ent;
	struct wr_timestamp ts;

	/* Verify that the size is correct */
	if (msg_i->header->len * 4 != sizeof(struct wrtd_trigger_entry))
		return -EINVAL;

	/* Copy the trigger instance and ack the message */
	memcpy(&ent, msg_i->payload, msg_i->header->len * 4);

	/* trigger entity ts from the host contains the delay.
	   So add to it the current time */
	struct wr_timestamp ts_orig;

	ts_orig.seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
	ts_orig.ticks = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);

	ts.seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
	ts.ticks = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);
	ts.frac = 0;
	ts_add(&ent.ts, &ts);


	if( ent.ts.seconds > ts_orig.seconds + 1000 )
	{
		pr_error("Warning: suspiciously long software trigger delay %"PRId64" tai-secs %"PRId64,
			 ent.ts.seconds, ts_orig.seconds);
	}

	/* Send trigger */
	send_trigger(&ent);
	wrtd_in_trigger_log(WRTD_LOG_SENT, 0, NULL, &ent);

	return 0;
}

void wr_enable_lock(int enable)
{
	dp_writel(TDC_CTRL_DIS_ACQ, BASE_DP_TDC_REGS + TDC_REG_CTRL);
	if (enable)
		dp_writel(TDC_WR_CTRL_ENABLE, BASE_DP_TDC_REGS + TDC_REG_WR_CTRL);
	else
		dp_writel(0, BASE_DP_TDC_REGS + TDC_REG_WR_CTRL);
	dp_writel(TDC_CTRL_EN_ACQ, BASE_DP_TDC_REGS + TDC_REG_CTRL);
}

/**
 * It updates the White-Rabbit link status
 */
void wr_update_link(void)
{
	int i;

	if (is_simulation) {
		wr_state = WR_LINK_SYNCED;
		return;
	}

	switch (wr_state) {
	case WR_LINK_OFFLINE:
		for(i = 0; i < TDC_NUM_CHANNELS; i++)
			wrtd_in_channels[i].config.flags |= WRTD_NO_WR;
		if (wr_link_up())
			wr_state = WR_LINK_ONLINE;
		break;
	case WR_LINK_ONLINE:
		if (wr_time_ready()) {
			wr_state = WR_LINK_SYNCING;
			wr_enable_lock(1);
		}
		break;
	case WR_LINK_SYNCING:
		if (wr_time_locked()) {
			pr_debug("rt-tdc: WR synced, waiting for TDC plumbing to catch up...\n\r");
			wr_state = WR_LINK_TDC_WAIT;
			tai_start = lr_readl(MT_CPU_LR_REG_TAI_SEC);
		}
		break;
	case WR_LINK_TDC_WAIT:
		if (lr_readl(MT_CPU_LR_REG_TAI_SEC) == (tai_start + 4)) {
			pr_debug("rt-tdc: WR TDC synced\n\r");
			wr_state = WR_LINK_SYNCED;
		}
		break;
	case WR_LINK_SYNCED:
		for(i = 0; i < TDC_NUM_CHANNELS; i++)
			wrtd_in_channels[i].config.flags &= ~WRTD_NO_WR;
		break;
	}

	if ( wr_state != WR_LINK_OFFLINE && !wr_link_up() ) {
		pr_error("rt-tdc: WR sync lost\n\r");
		wr_state = WR_LINK_OFFLINE;
		wr_enable_lock(0);
	}
}


/**
 * Data structures to export to host system
 */
static struct trtl_fw_buffer wrtd_in_buffers[] = {
	[IN_STRUCT_DEVICE] = {
		.buf = &wrtd_in_dev,
		.len = sizeof(struct wrtd_in),
	},
	[IN_STRUCT_CHAN_0] = {
		.buf = &wrtd_in_channels[0],
		.len = sizeof(struct wrtd_in_channel),
	},
	[IN_STRUCT_CHAN_1] = {
		.buf = &wrtd_in_channels[1],
		.len = sizeof(struct wrtd_in_channel),
	},
	[IN_STRUCT_CHAN_2] = {
		.buf = &wrtd_in_channels[2],
		.len = sizeof(struct wrtd_in_channel),
	},
	[IN_STRUCT_CHAN_3] = {
		.buf = &wrtd_in_channels[3],
		.len = sizeof(struct wrtd_in_channel),
	},
	[IN_STRUCT_CHAN_4] = {
		.buf = &wrtd_in_channels[4],
		.len = sizeof(struct wrtd_in_channel),
	},
};

static struct trtl_fw_variable wrtd_in_variables[] = {
	[IN_VAR_CHAN_ENABLE] = {
		.addr = TRTL_ADDR_DP(BASE_DP_TDC_DIRECT + DR_REG_CHAN_ENABLE),
		.mask = 0x2F,
		.offset = 16,
	},
	[IN_VAR_DEVICE_TIME_S] = {
		.addr = TRTL_ADDR_LR(MT_CPU_LR_REG_TAI_SEC),
		.mask = 0xFFFFFFFF,
		.offset = 0,
	},
	[IN_VAR_DEVICE_TIME_T] = {
		.addr = TRTL_ADDR_LR(MT_CPU_LR_REG_TAI_CYCLES),
		.mask = 0xFFFFFFFF,
		.offset = 0,
	},
	[IN_VAR_DEVICE_SENT_PACK] = {
		.addr = &sent_packets,
		.mask = 0xFFFFFFFF,
		.offset = 0,
	},
	[IN_VAR_DEVICE_DEAD_TIME] = {
		.addr = TRTL_ADDR_DP(BASE_DP_TDC_DIRECT + DR_REG_DEAD_TIME),
		.mask = 0xFFFFFFFF,
		.offset = 0,
	},
	[IN_VAR_DEVICE_CHAN_ENABLE] = {
		.addr = TRTL_ADDR_DP(BASE_DP_TDC_DIRECT + DR_REG_CHAN_ENABLE),
		.mask = 0x1F,
		.offset = 0,
	},
};

static trtl_fw_action_t *wrtd_in_actions[] = {
	[WRTD_IN_ACTION_SW_TRIG] = wrtd_in_trigger_sw,
};

static int wrtd_i_sim_init(void)
{
	int i;

	for(i = 0; i < TDC_NUM_CHANNELS; i++) {
		wrtd_in_channels[i].config.log_level = WRTD_LOG_ALL;
		wrtd_in_channels[i].config.id.system = 0;
		wrtd_in_channels[i].config.id.source_port = i;
		wrtd_in_channels[i].config.id.trigger = i;
		wrtd_in_channels[i].config.flags =
		  WRTD_ENABLED | WRTD_ARMED | WRTD_TRIGGER_ASSIGNED;
	}

	dp_writel(0x40, BASE_DP_TDC_DIRECT + DR_REG_DEAD_TIME);
	dp_writel(0x1f, BASE_DP_TDC_DIRECT + DR_REG_CHAN_ENABLE);

	return 0;
}

static int wrtd_i_init(void)
{
	struct trtl_ep_eth_address addr = {
		.type = TRTL_EP_FRAME_UDP,
		.dst_mac = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF},
		.dst_ip = 0xFFFFFFFF, /* broadcast on 255.255.255.255 */
		.src_ip = 0xC0A85A11,
		.dst_port = WRTD_OUT_PORT,
		.src_port = WRTD_IN_PORT,
		.ethertype = 0x800, /* IPv4 */
		.filter = 0 /* only the TX path of the endpoint is used */
	};
	int i;

	sent_packets = 0;
	coalesce_count = 0;
	wr_state = WR_LINK_OFFLINE;
	tai_start = 0;

	if (!is_simulation)
		tdc_init();

	wr_enable_lock(0);

	loop_queue_init();

	/* Set up channel states to safe default values */
	memset(wrtd_in_channels, 0,
	       sizeof(struct wrtd_in_channel) * TDC_NUM_CHANNELS);
	for(i = 0; i < TDC_NUM_CHANNELS; i++) {
		wrtd_in_channels[i].n = i;
		wrtd_in_channels[i].config.mode = WRTD_TRIGGER_MODE_AUTO;
	}

	rmq_bind_out(WRTD_IN_RMQ, TRTL_EP_ETH, &addr);

	/* Skip WR sync and automatically generate some events when simulating */
	if (is_simulation)
		wrtd_i_sim_init();

	pr_debug("rt-input firmware initialized.\n\r");

	return 0;
}

static int wrtd_i_main(void)
{
	while (1) {
		do_input();
		trtl_fw_mq_action_dispatch(TRTL_HMQ, WRTD_IN_HMQ);
		wr_update_link();
	}

	return 0;
}

struct trtl_fw_application app = {
	.name = "wrtd-input",
	.version = {
		.rt_id = WRTD_IN_RT_ID,
		.rt_version = RT_VERSION(2, 0),
		.git_version = GIT_VERSION
	},

	.buffers = wrtd_in_buffers,
	.n_buffers = ARRAY_SIZE(wrtd_in_buffers),

	.variables = wrtd_in_variables,
	.n_variables = ARRAY_SIZE(wrtd_in_variables),

	.actions = wrtd_in_actions,
	.n_actions = ARRAY_SIZE(wrtd_in_actions),

	.init = wrtd_i_init,
	.main = wrtd_i_main,
	.exit = NULL,
};
